#include <X11/Xlib.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <X11/keysym.h>
#include <X11/Xutil.h>
#include "Globals.h"



//Drawing functions
void DrawButtons(Display*, struct Manager,  int);
void DrawFit(Display*,struct Manager,double[],double[],double[],int,int,int);
void DrawLabel(Display*,struct Manager,char*);
int spettro(Display*, struct Manager, double[], double[], int);

//Load functions
int loadspettro(char*, double*, double*, int);
int datacount(char*);

//Fitting function
extern void bestfit_(double*, int*, double*, double*, int*);

//Extern functions
void InitPixmap(Display*,Window,Manager*);
void FreePixmap(Display*, Manager);
double xconversione(Display*, Manager, double[], int, int, double, int);
double yconversione(Display*, Manager, double[], int, int, double, int);
int xdiretta(Display*, Manager, double[], int, double);
double xinversa(Display*, Manager, double[], int, int);
int ydiretta(Display*, Manager, double[], int, double);
double yinversa(Display*, Manager, double[], int, int);


void writeOutput(char* input,double* params,int nLines);
int main(int argc, char** argv)
{
  /* Dichiaro le variabili */
  /* Xlib variables */
  Display *display;
  Window window;
  int screen_num;
  int screen_width;
  int screen_height;
  int border_width;
  unsigned long background;
  unsigned long border;
  int x_window;
  int y_window;
  XEvent event, event2;
  /* End Xlib variables */
  
  Manager man;
  int ndata;
  int done=0,selection=-1;
  
  double *x, *y;
  
  if (argc<2){
	printf("This fitting software needs the input file as parameter.");
	printf("Please pass it from command line");
	return -1;
  }
  if(strcmp(argv[1],"-h")==0)
  {
	printf("MCMC-like fitting software\n Needed parameter: spectrum file name\n");
	printf("The spectrum file must be an ascii file with two columns without header:\n");
	printf("Column 1: lambda\nColumn 2: flux/nu*flux\n");
	return 0;
  }
  ndata = datacount(argv[1]);
  if (ndata == -1) return -1;
  int i;
  /* Allocazione */
  x = malloc(ndata*sizeof(double));
  y = malloc(ndata*sizeof(double));
  
  /* Carico i dati */
  loadspettro(argv[1], x, y, ndata);
  
  /* Assegnamenti */
  display = XOpenDisplay(NULL);
  screen_num = DefaultScreen(display);
  screen_width = DisplayWidth(display, screen_num);
  screen_height = DisplayHeight(display, screen_num); 
  
  background = WhitePixel(display, screen_num);
  border = BlackPixel(display, screen_num);
  border_width = 2;
  x_window = 1;
  y_window = 1;
  
  /* Creo la finestra */
  window = XCreateSimpleWindow(display, RootWindow(display, screen_num), 
  				x_window, y_window, screen_width, screen_height, 
				border_width, border, background);
  /*window = XCreateSimpleWindow(display, RootWindow(display, screen_num), 
  				x_window, y_window, 640,480, 
				border_width, border, background);*/
  
  
  /* Creo la "man.penna" con cui disegnero' */
  
  
  /* Gestione chiusura finestra */
  Atom wmDeleteMessage = XInternAtom(display, "WM_DELETE_WINDOW", False);
  XSetWMProtocols(display, window, &wmDeleteMessage, 1);
  
  XColor red_col,green_col,blue_col;
  Colormap colormap = DefaultColormap(display, screen_num);
  
  XGCValues values;
  values.foreground = BlackPixel(display, screen_num);
  values.background= WhitePixel(display,screen_num);
  values.line_width = 1;
  values.line_style = LineSolid;
  
  char value[8],frase[200];
  /* Creo il rosso */
  strcpy(value,"#FF0000");
  XParseColor(display, colormap, value, &red_col);
  XAllocColor(display, colormap, &red_col);
  /* Creo il verde */
  strcpy(value,"#00C000");
  XParseColor(display, colormap, value, &green_col);
  XAllocColor(display, colormap, &green_col);
  strcpy(value,"#0000FF");
  /* Creo il blu */
  XParseColor(display, colormap, value, &blue_col);
  XAllocColor(display, colormap, &blue_col);
  
  char labels[3][80]={{"Select the continuum flux"},
		   {"Select the line's peak"},
		   {"Select the line's FWHM"}};
  


  /*Variabili del ciclo di acquisizione dei parametri iniziali per il fit*/
  int counter=0;
  int x1, y1, x2, y2;
  int px[300];
  int py[300];
  double a[600];
  int vector[600];
  int j=0;
  
  int broad=0;
  /* event loop */
  
  InitPixmap(display,window,&man);
  XSetBackground(display,man.back,WhitePixel(display,screen_num));
  XSetForeground(display,man.back,BlackPixel(display,screen_num));
  XSetForeground(display,man.pen,BlackPixel(display,screen_num));
  XSetForeground(display,man.red,red_col.pixel);  
  XSetForeground(display,man.green,green_col.pixel);
  XSetForeground(display,man.blue,blue_col.pixel);
  ClearPixmap(display,man);
  //XSync(display,0);    
  /* Comunico al display che tipo di eventi volgliamo registrare */
  XSelectInput(display, window, ButtonPressMask|ButtonReleaseMask|StructureNotifyMask|ExposureMask|KeyPressMask|KeyReleaseMask);
  /* Rendo visibile la finestra*/
  XMapWindow(display, window);
  int looper=1;
  
	    
  
  while (looper)
    {
        XNextEvent(display, &event);
        /* disegna o ridisegna la finestra */
        switch(event.type)
        {
	  
	  case ClientMessage: /* Chiusura finestra */
	    if(event.xclient.data.l[0]==(unsigned int) wmDeleteMessage){
	     	FreePixmap(display,man);
	     	XCloseDisplay(display);
	        looper=0;
	    }
	    break;
	  case ConfigureNotify:
	    {
	    FreePixmap(display,man);
	    InitPixmap(display,window,&man);
	    XSetForeground(display,man.back,WhitePixel(display,screen_num));
            XSetForeground(display,man.pen,BlackPixel(display,screen_num));
            XSetForeground(display,man.red,red_col.pixel);  
            XSetForeground(display,man.green,green_col.pixel);
            XSetForeground(display,man.blue,blue_col.pixel);
	    ClearPixmap(display,man);
	    spettro(display, man, x, y,  ndata);
	    //XClearArea(display,window,0,0,0,0,0);
	    
	    }
	    break;
	  case Expose:
	    XSync(display,0);
	    
	    spettro(display, man, x, y,  ndata);
	    if(done) {
	     DrawButtons(display,man,selection);
	     DrawFit(display,man,x,y,a,ndata,j,selection);
	    } 
	    else
	    {
	     DrawLabel(display,man,labels[counter]);
	    }
	    //XFlush(display);
	    XCopyArea(display, man.map, window, man.pen, 0, 0, man.width,man.height, 0,0); 
	    //XFlush(display);
	    break; 
	    
	  /* prendo i valori x e y corrispondenti al click e al suo rilascio */
	  case ButtonPress:
	    if(!done){
	     if(counter==0) /*Seleziona il continuo*/
	     {
		y1 = event.xbutton.y;
		py[counter] = y1;
		vector[j] = py[0];
		a[j] = yinversa(display, man, y, ndata, vector[j]);
		printf("Input: continuum flux (j=%d) \t a[%d]=%e\n", j,j, a[j]);
		/* Assegno un paramentro iniziale per la legge di potenza (esponente) */
		j=j+1;
		a[j] = -0.01;
		printf("Input: continuum exponent (j=%d) \t a[%d]=%f\n", j,j, a[j]);
	     }
	     else if(counter==1) /*Seleziona il picco*/
	     {
		x1 = event.xbutton.x;
		y1 = event.xbutton.y; 
		px[counter] = x1;
		py[counter] = y1;
		vector[j] = px[1];
		a[j] = xinversa(display, man, x, ndata, vector[j]);
		printf("Input: line peak lambda (j=%d) \t a[%d]=%f\n", j,j, a[j]);
		j=j+1;
		vector[j] = abs(py[1]);
		a[j] = yinversa(display, man, y, ndata, vector[j]);
		printf("Input: line peak flux (j=%d) \t a[%d]=%e\n", j,j, a[j]);
	     }
	     else if(counter==2)
	     {
		x1 = event.xbutton.x;
		j--;
	     }
	     j++;
	    }
	    else
	    {
	     if(event.xbutton.y>FRAME/2-20 && event.xbutton.y<FRAME/2){
	      int delta=(man.width-FRAME-event.xbutton.x);
	      selection=-1;
	      if(delta<360 && delta>260) 
	      	selection=0;
	      else if(delta<240 && delta>140)
	        selection=1;
	      else if(delta<120 && delta>20)
	        selection=2;
	      if(selection>=0) {
	        ClearPixmap(display,man);
	      	XClearArea(display,window,0,0,0,0,1);
	      }
	     }
	    }
	    break;
	  case ButtonRelease:
	    if(!done){
	     if(counter==0) /*Continuo*/
	     {
	        XDrawRectangle(display, man.map, man.red, event.xbutton.x-3,
			 event.xbutton.y-3,5,5);
		counter++;
	     }
	     else if(counter==1) /*Picco*/
	     {
	        XDrawRectangle(display, man.map, man.red, event.xbutton.x-3,
			 event.xbutton.y-3,5,5);
		counter++;
	     }
	     else if(counter==2) /*Larghezza*/
	     {
	 	x2 = event.xbutton.x;
	        XDrawLine(display, man.map, man.red, x1,event.xbutton.y,x2,
			  event.xbutton.y);
	 	double x1conv, x2conv;
	 	x1conv = xinversa(display, man, x, ndata, x1);
	 	x2conv = xinversa(display, man, x, ndata, x2);
	 	px[counter] = abs(x2conv-x1conv); /*FWHM*/
	 	if(px[counter]==0) px[counter]=1;
		a[j] = 2*pow((px[2]/2.355),2);/*denominatore dell'esp della gaussiana*/
	 	printf("Input: line FWHM (j=%d) \t a[%d]=%e\n", j,j, a[j]);
		j++;
	 	counter--;
	     } 
	     int length = strlen(frase);
	     XClearArea(display, window, 85, 50, 800, 30,0);
	     XFillRectangle(display,man.map,man.back,85,50,800,30);
	     DrawLabel(display,man,labels[counter]);
	     XCopyArea(display, man.map, window, man.pen, 0, 0, man.width,man.height, 0,0); 
	     XFlush(display);
 	     break;
	    }
	  case KeyPress:
	    if(XLookupKeysym(&event.xkey, 0) == XK_Return && !done)
	    {
	      FILE* outname=fopen("outname","w");
	      fprintf(outname,"%s",argv[1]);
	      fclose(outname);
	      bestfit_(a, &j, x, y, &ndata);
	      system("rm outname");
	      done=1;
	      selection=0;
              ClearPixmap(display,man);	
	      XClearArea(display,window,0,0,0,0,1);
	      XFlush(display);
	      break;
	    }
	    /* Chiudo la connessione premendo q */ 
	    else if(XLookupKeysym(&event.xkey, 0) == XK_q)
	    {
	      XFreePixmap(display,man.map);
	      XCloseDisplay(display);
	      looper=0;
	    }
	    
	}
    }
    
    if(done) writeOutput(argv[1],a,(j-2)/3);
    return 0;
} 







/*FUNZIONI*/

void writeOutput(char* input,double* params,int nLines)
{
  char output[160];
  char* token;//=strtok(input,".");
  char * substr;
  substr=input;
  int pCount=0;
  while((substr=strstr(substr,"/"))!=NULL)
  {
	pCount++;
	substr++;
  }
  token=strtok(input,"/");
  pCount--;
  for(pCount;pCount>0;pCount--)
  	token=strtok(NULL,"/");
  token=strtok(NULL,".");
  int i=0;
  sprintf(output,"par_%s.dat",token);
  //puts(output);
  FILE* outf=fopen(output,"w");
  fprintf(outf,"Latest fit parameters\n- Continuum parameters:\n");
  fprintf(outf,"---- Flux - Exp ----\n");
  fprintf(outf,"%.4g\t %.4g\n",params[0],params[1]);
  fprintf(outf,"- Lines:%d\n",nLines);
  fprintf(outf,"---- Centre - Flux - FWHM ----\n");
  for(i=1;i<=nLines;i++)
  {
    fprintf(outf,"%.2f\t %.4g\t %.4f\n",params[3*i-1],params[3*i],params[3*i+1]);
  }
  fclose(outf);
}

void DrawLabel(Display* display,struct Manager man,char* frase)
{
 XDrawString(display, man.map, man.pen, 90,70 , frase, strlen(frase));
}

void DrawFit(Display* display,struct Manager man,double x[],double y[],double a[],
		int ndata,int nparam,int selection){
	char frase[] = "Resulting fit";
	int length = strlen(frase);
	//ClearPixmap(display,man);
	XFillRectangle(display,man.map,man.back,85,50,800,30);
	     
	//XClearArea(display, window, 85, 50, 800, 30,0);
	XDrawString(display, man.map, man.pen, 90,70 , frase, length);
		
	int i=0,k=0,l=0;
	switch(selection)
	{
	 case 0:
	 	for(i=0; i<ndata-1; i++)
	      	{
		  double fit[ndata];
		  fit[i]=a[0]*pow((x[i]/6000.),a[1]);
		  /* Questo ciclo, per ogni i, assegna tutti i parametri */
		  k=0;
		  int nlines=(nparam-2)/3;

		  for(k=0; k<nlines; k++)
		  {
		      fit[i]=fit[i]+a[3+3*k]*exp(-pow((x[i]-a[2+3*k]),2)/a[4+k*3]);
		  } 

		  if(i>0)
		  {
		      int xin = xdiretta(display, man, x, ndata, x[i-1]);
		      int xfin = xdiretta(display, man, x, ndata, x[i]);
		      int yin = ydiretta(display, man, y, ndata, fit[i-1]);
		      int yfin = ydiretta(display, man, y, ndata, fit[i]);
		      XDrawLine(display, man.map, man.red, xin, yin, xfin, yfin);
		  }
	      
	      	}
		break;
	 case 1: /*Narrow Line*/
	 	k=4;
	      	while(k<nparam)
		{
		   if(a[k]<200) 
			{
			  printf("Narrow: a[%d]\n centre=%f\n FWHM=%f\n",k, a[k-2], sqrt(a[k]/2)*2.355);
			  for(i=0;i<ndata-1;i++)
			    {
			      double gauss[ndata];
			      l=2;
			      
			      while(l<nparam)
			      {
				
				gauss[i]=a[0]*pow((x[i]/6000.),a[1])+a[k-1]*exp(-pow((x[i]-a[k-2]),2)/a[k]);
				l=l+3;
			      }
			      
			    if(i>0)
			      {
				int xgauss1 = xdiretta(display, man, x, ndata, x[i-1]);
				int ygauss1 = ydiretta(display, man, y, ndata, gauss[i-1]);
				int xgauss2 = xdiretta(display, man, x, ndata, x[i]);
				int ygauss2 = ydiretta(display, man, y, ndata, gauss[i]);

				XDrawLine(display, man.map, man.blue, xgauss1, ygauss1, xgauss2, ygauss2);
			      }
			    }

			}
			k=k+3;
		}
	 	break;
	 case 2: /*Broad Line*/
	 	k=4;
	      	while(k<nparam)
		{
		    if(a[k]>200) 
		    {
			  printf("Broad: a[%d]\n centre=%f\n FWHM=%f\n",k, a[k-2], sqrt(a[k]/2)*2.355);
			  for(i=0;i<ndata-1;i++)
			  {
			    double gauss[ndata];
			    gauss[i]=a[0]*pow((x[i]/6000.),a[1])+a[k-1]*exp(-pow((x[i]-a[k-2]),2)/a[k]);
			    if(i>0)
			      {
				int xgauss1 = xdiretta(display, man, x, ndata, x[i-1]);
				int ygauss1 = ydiretta(display, man, y, ndata, gauss[i-1]);
				int xgauss2 = xdiretta(display, man, x, ndata, x[i]);
				int ygauss2 = ydiretta(display, man, y, ndata, gauss[i]);

				XDrawLine(display, man.map, man.green, xgauss1, ygauss1, xgauss2, ygauss2);
			      }
			  }
		    }
		    k+=3;
		}
	 	break;
	}
}



void DrawButtons(Display *display,struct Manager man,int selection)
{
  XFillRectangle(display, man.map, man.back, man.width-FRAME-100, FRAME/2-20, 100, 20);
  XFillRectangle(display, man.map, man.back, man.width-FRAME-220, FRAME/2-20, 100, 20);
  XFillRectangle(display, man.map, man.back, man.width-FRAME-340, FRAME/2-20, 100, 20);
  XDrawRectangle(display, man.map, man.pen, man.width-FRAME-100, FRAME/2-20, 100, 20);
  XDrawRectangle(display, man.map, man.pen, man.width-FRAME-220, FRAME/2-20, 100, 20);
  XDrawRectangle(display, man.map, man.pen, man.width-FRAME-340, FRAME/2-20, 100, 20);
  int x=man.width-FRAME-120*(3-selection)+21;
  XFillRectangle(display,man.map,man.green,x,FRAME/2-19,99,19);
  char label[30];
  strcpy(label,"Whole Fit");
  XDrawString(display, man.map, man.pen, man.width-FRAME-317, FRAME/2-5, label, strlen(label));
  strcpy(label,"Narrow Lines");
  XDrawString(display, man.map, man.pen, man.width-FRAME-205, FRAME/2-5, label, strlen(label));
  strcpy(label,"Broad Lines");
  XDrawString(display, man.map, man.pen, man.width-FRAME-82, FRAME/2-5, label, strlen(label));
}


int spettro(Display *display, Manager man, double *x, double *y, int ndata)
{
  XDrawRectangle(display, man.map, man.pen, FRAME, FRAME, man.width-2*FRAME, man.height-2*FRAME);
  double ymax = y[0];
  double ymin = y[0];
  int i;
  for (i=1; i<ndata; i++)
  {
    if(ymax<y[i]) ymax=y[i];
      
    if(ymin>y[i]) ymin=y[i];
    
  }
  
  double yfac = (man.height-2*FRAME)/(ymax-ymin)*YSCALE; 
  for (i = 0; i < ndata-1; i++)
  {
    int xin, xfin, yin, yfin; /*coordinate in pixel*/
    xin = xdiretta(display, man, x, ndata, x[i]);
    xfin = xdiretta(display, man, x, ndata, x[i+1]);
    yin = ydiretta(display, man, y, ndata, y[i]);
    yfin = ydiretta(display, man, y, ndata, y[i+1]);
    XDrawLine(display, man.map, man.pen, xin, yin, xfin, yfin);
  }
  
   /*Disegno la scala negli assi*/
   
   int xstep = (x[ndata-1]-x[0])/20;
      
   int yaxis, xright, xleft;
   double xstart, ystart;
   xstart = (int)x[0] +1;
   yaxis = man.height-FRAME ;
   xleft = FRAME -5;
   xright = FRAME +5;
   
   /*Definisco l'y minimo dell'asse in flusso*/
   double yzero = yinversa(display, man, y, ndata, yaxis);
   /*Definisco l'y massimo dell'asse in flusso*/
   double ysup = yinversa(display, man, y, ndata, FRAME);
   ystart = yzero; 
   
   double ystep = (ysup-yzero)/10;
   
   while (xstart < x[ndata-1])
	    {
	      int xline = (int)xdiretta(display, man, x, ndata, xstart);
	      XDrawLine(display, man.map, man.pen, xline, yaxis+5, xline, yaxis-5);
	      int number = (int)xstart;
	      char xnumber[10];
	      sprintf(xnumber, "%d", number);
	      int length = strlen(xnumber);
	      XDrawString(display, man.map, man.pen, xline-10, yaxis+20, xnumber , length);
	      
	      xstart = xstart + xstep;
	    }
	  
   while (ystart < ysup)
	    {
	      double conversion = (ystart-yzero)*yfac;
	      int yline = man.height-FRAME - (int)conversion;	
	      XDrawLine(display, man.map, man.pen, xleft, yline, xright, yline);
	      double number = ystart;
	      char ynumber[30];
	      sprintf(ynumber, "%7.3e", number);
	      int length = strlen(ynumber);
	          
	      
	      XDrawString(display, man.map, man.pen, FRAME-80, yline+4, ynumber , length);
	      
	      ystart = ystart + ystep;
	    }
	    
  return 1;
  
}

int loadspettro(char* filename, double* x, double* y, int nn)
{
  FILE* file;
  file = fopen(filename, "r");

  int i;

  for (i = 0; i < nn; i++)
  {
    fscanf(file, "%lf%lg", &x[i], &y[i]);
  }
  fclose(file);
  return 1;
}

int datacount(char* filename)
{
  int nn=0;
  FILE* file;
  file = fopen(filename, "r");
  if (file == NULL)
  {
    printf("Input file does not exist!\n");
    return -1;
  }
  char riga[81];
  while (!feof(file)) /*feof e' una funzione che indica la fine del file*/
  {
    fgets(riga, 80, file);
    nn=nn+1;
  }
  nn=nn-1;
  fclose(file);
  return nn;
 }
 
